@@ -14,14 +14,14 @@
 
 """Execute Transactions Spec tests."""
 
-import functools
 import os
 import sys
-import time
 import threading
+import time
 
 sys.path[0:0] = [""]
 
+from bson.son import SON
 from bson.objectid import ObjectId
 
 from pymongo.errors import (ConnectionFailure,
@@ -37,24 +37,29 @@ from pymongo.monitoring import (ConnectionCheckedInEvent,
                                 ConnectionCreatedEvent,
                                 ConnectionReadyEvent,
                                 PoolCreatedEvent,
+                                PoolReadyEvent,
                                 PoolClearedEvent,
                                 PoolClosedEvent)
 from pymongo.read_preferences import ReadPreference
-from pymongo.pool import _PoolClosedError
+from pymongo.pool import _PoolClosedError, PoolState
+from pymongo.topology_description import updated_topology_description
 
-from test import (IntegrationTest,
+from test import (client_knobs,
+                  IntegrationTest,
                   unittest)
 from test.utils import (camel_to_snake,
                         client_context,
                         CMAPListener,
                         get_pool,
                         get_pools,
+                        OvertCommandListener,
                         rs_or_single_client,
                         single_client,
                         single_client_noauth,
                         TestCreator,
                         wait_until)
 from test.utils_spec_runner import SpecRunnerThread
+from test.pymongo_mocks import DummyMonitor
 
 
 OBJECT_TYPES = {
@@ -67,6 +72,7 @@ OBJECT_TYPES = {
     'ConnectionReady': ConnectionReadyEvent,
     'ConnectionCheckOutStarted': ConnectionCheckOutStartedEvent,
     'ConnectionPoolCreated': PoolCreatedEvent,
+    'ConnectionPoolReady': PoolReadyEvent,
     'ConnectionPoolCleared': PoolClearedEvent,
     'ConnectionPoolClosed': PoolClosedEvent,
     # Error types.
@@ -101,13 +107,15 @@ class TestCMAP(IntegrationTest):
         thread.join()
         if thread.exc:
             raise thread.exc
+        self.assertFalse(thread.ops)
 
     def wait_for_event(self, op):
         """Run the 'waitForEvent' operation."""
         event = OBJECT_TYPES[op['event']]
         count = op['count']
+        timeout = op.get('timeout', 10000) / 1000.0
         wait_until(lambda: self.listener.event_count(event) >= count,
-                   'find %s %s event(s)' % (count, event))
+                   'find %s %s event(s)' % (count, event), timeout=timeout)
 
     def check_out(self, op):
         """Run the 'checkOut' operation."""
@@ -126,6 +134,10 @@ class TestCMAP(IntegrationTest):
         sock_info = self.labels[label]
         self.pool.return_socket(sock_info)
 
+    def ready(self, op):
+        """Run the 'ready' operation."""
+        self.pool.ready()
+
     def clear(self, op):
         """Run the 'clear' operation."""
         self.pool.reset()
@@ -177,6 +189,8 @@ class TestCMAP(IntegrationTest):
         """Check the events of a test."""
         actual_events = self.actual_events(ignore)
         for actual, expected in zip(actual_events, events):
+            self.logs.append('Checking event actual: %r vs expected: %r' % (
+                actual, expected))
             self.check_event(actual, expected)
 
         if len(events) > len(actual_events):
@@ -187,18 +201,62 @@ class TestCMAP(IntegrationTest):
         self.check_object(actual, expected)
         self.assertIn(message, str(actual))
 
+    def _set_fail_point(self, client, command_args):
+        cmd = SON([('configureFailPoint', 'failCommand')])
+        cmd.update(command_args)
+        client.admin.command(cmd)
+
+    def set_fail_point(self, command_args):
+        if not client_context.supports_failCommand_fail_point:
+            self.skipTest('failCommand fail point must be supported')
+        self._set_fail_point(self.client, command_args)
+
     def run_scenario(self, scenario_def, test):
         """Run a CMAP spec test."""
+        self.logs = []
         self.assertEqual(scenario_def['version'], 1)
-        self.assertEqual(scenario_def['style'], 'unit')
+        self.assertIn(scenario_def['style'], ['unit', 'integration'])
         self.listener = CMAPListener()
         self._ops = []
 
+        # Configure the fail point before creating the client.
+        if 'failPoint' in test:
+            fp = test['failPoint']
+            self.set_fail_point(fp)
+            self.addCleanup(self.set_fail_point, {
+                'configureFailPoint': fp['configureFailPoint'], 'mode': 'off'})
+
         opts = test['poolOptions'].copy()
         opts['event_listeners'] = [self.listener]
-        client = single_client(**opts)
+        opts['_monitor_class'] = DummyMonitor
+        opts['connect'] = False
+        # Support backgroundThreadIntervalMS, default to 50ms.
+        interval = opts.pop('backgroundThreadIntervalMS', 50)
+        if interval < 0:
+            kill_cursor_frequency = 99999999
+        else:
+            kill_cursor_frequency = interval/1000.0
+        with client_knobs(kill_cursor_frequency=kill_cursor_frequency,
+                          min_heartbeat_interval=.05):
+            client = single_client(**opts)
+            # Update the SD to a known type because the DummyMonitor will not.
+            # Note we cannot simply call topology.on_change because that would
+            # internally call pool.ready() which introduces unexpected
+            # PoolReadyEvents. Instead, update the initial state before
+            # opening the Topology.
+            td = client_context.client._topology.description
+            sd = td.server_descriptions()[(client_context.host,
+                                           client_context.port)]
+            client._topology._description = updated_topology_description(
+                client._topology._description, sd)
+            # When backgroundThreadIntervalMS is negative we do not start the
+            # background thread to ensure it never runs.
+            if interval < 0:
+                client._topology.open()
+            else:
+                client._get_topology()
         self.addCleanup(client.close)
-        self.pool = get_pool(client)
+        self.pool = list(client._topology._servers.values())[0].pool
 
         # Map of target names to Thread objects.
         self.targets = dict()
@@ -226,8 +284,7 @@ class TestCMAP(IntegrationTest):
             self.check_events(test['events'], test['ignore'])
         except Exception:
             # Print the events after a test failure.
-            print()
-            print('Failed test: %r' % (test['description'],))
+            print('\nFailed test: %r' % (test['description'],))
             print('Operations:')
             for op in self._ops:
                 print(op)
@@ -238,6 +295,9 @@ class TestCMAP(IntegrationTest):
             print('Events:')
             for event in self.listener.events:
                 print(event)
+            print('Log:')
+            for log in self.logs:
+                print(log)
             raise
 
     POOL_OPTIONS = {
@@ -323,13 +383,14 @@ class TestCMAP(IntegrationTest):
             client.admin.command('ping')
 
         self.assertIsInstance(listener.events[0], PoolCreatedEvent)
-        self.assertIsInstance(listener.events[1],
-                              ConnectionCheckOutStartedEvent)
+        self.assertIsInstance(listener.events[1], PoolReadyEvent)
         self.assertIsInstance(listener.events[2],
+                              ConnectionCheckOutStartedEvent)
+        self.assertIsInstance(listener.events[3],
                               ConnectionCheckOutFailedEvent)
-        self.assertIsInstance(listener.events[3], PoolClearedEvent)
+        self.assertIsInstance(listener.events[4], PoolClearedEvent)
 
-        failed_event = listener.events[2]
+        failed_event = listener.events[3]
         self.assertEqual(
             failed_event.reason, ConnectionCheckOutFailedReason.CONN_ERROR)
 
@@ -345,17 +406,16 @@ class TestCMAP(IntegrationTest):
             client.admin.command('ping')
 
         self.assertIsInstance(listener.events[0], PoolCreatedEvent)
-        self.assertIsInstance(listener.events[1],
+        self.assertIsInstance(listener.events[1], PoolReadyEvent)
+        self.assertIsInstance(listener.events[2],
                               ConnectionCheckOutStartedEvent)
-        self.assertIsInstance(listener.events[2], ConnectionCreatedEvent)
+        self.assertIsInstance(listener.events[3], ConnectionCreatedEvent)
         # Error happens here.
-        self.assertIsInstance(listener.events[3], ConnectionClosedEvent)
-        self.assertIsInstance(listener.events[4],
+        self.assertIsInstance(listener.events[4], ConnectionClosedEvent)
+        self.assertIsInstance(listener.events[5],
                               ConnectionCheckOutFailedEvent)
-
-        failed_event = listener.events[4]
-        self.assertEqual(
-            failed_event.reason, ConnectionCheckOutFailedReason.CONN_ERROR)
+        self.assertEqual(listener.events[5].reason,
+                         ConnectionCheckOutFailedReason.CONN_ERROR)
 
     #
     # Extra non-spec tests
@@ -381,6 +441,20 @@ class TestCMAP(IntegrationTest):
         self.assertRepr(PoolClearedEvent(host, service_id=ObjectId()))
         self.assertRepr(PoolClosedEvent(host))
 
+    def test_close_leaves_pool_unpaused(self):
+        # Needed until we implement PYTHON-2463. This test is related to
+        # test_threads.TestThreads.test_client_disconnect
+        listener = CMAPListener()
+        client = single_client(event_listeners=[listener])
+        client.admin.command('ping')
+        pool = get_pool(client)
+        client.close()
+        self.assertEqual(1, listener.event_count(PoolClearedEvent))
+        self.assertEqual(PoolState.READY, pool.state)
+        # Checking out a connection should succeed
+        with pool.get_socket({}):
+            pass
+
 
 def create_test(scenario_def, test, name):
     def run_scenario(self):
